#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Mar  7 20:38:55 2020

@author: marisaeisenberg
"""

import pycxsimulator
from pylab import *
from copy import deepcopy
from pyDOE import lhs
import networkx as nx
import seaborn as sns
import numpy.matlib as npm



'''
Model Setup - SIR model with tie-breaking and regrowth
'''

# Parameters
p_i = 0.5 # infection probability per contact
p_r = 0.5 # recovery probability
p_0 = 0.5 # probability of breaking tie if infected neighbor
p_1 = 0.1 # probability of regrowing tie if non-infected

prev = []
lostedges = []


# Model functions
def initialize():
    global g, nextg, prev
    prev = []
    lostedges = []
    g = nx.karate_club_graph()
    # g = nx.barabasi_albert_graph(50,5)
    g.pos = nx.spring_layout(g)
    nx.set_node_attributes(g, 0, 'state')  # everyone starts off susceptible
    g.node[1]['state'] = 1                 # set one node to be infected (index case)
    nextg = g.copy()
    nextg.pos = g.pos
    prev.append(1/len(g.nodes)) # initial prevalence

def update():
    global g, nextg, prev
    curprev = 0
    nextg = deepcopy(g)
    for a in g.nodes:
        if g.node[a]['state'] == 0: # if susceptible
            nextg.node[a]['state'] = 0
            for b in g.neighbors(a):
                if g.node[b]['state'] == 1: # if neighbor b is infected
                    if random() < p_0:  # with probability p_0 cut them off at next time step
                        nextg.remove_edge(a, b)
                        lostedges.append([a,b])
                    if random() < p_i: # but we still let them potentially transmit now
                        nextg.node[a]['state'] = 1                    
        else: # if infected
            curprev += 1
            nextg.node[a]['state'] = 0 if random() < p_r else 1
    
    for i in range(len(lostedges)): # go through dropped edges and see which ones we can get back
        if g.node[lostedges[i][0]]['state'] == 0 and g.node[lostedges[i][1]]['state'] == 0:
            if random() < p_1:
                nextg.add_edge(lostedges[i][0],lostedges[i][1])
    
    prev.append(curprev/len(g.nodes()))
    g = deepcopy(nextg)
    
  
    
'''
Examine stochastic variability with default parameters
'''
   
# Run the model with the default parameters to examine the effect of stochastic variability
# Outputs: peak prevalence of the epidemic, total duration of the epidemic

timesteps = 100
numruns = 100
maxprev = []
durations = []
trajectories = np.zeros([timesteps,numruns])
for i in range(0,numruns):
    initialize()
    for j in range(1,timesteps):
        update()
    maxprev.append(max(prev))
    durations.append(sum(i > 0 for i in prev))
    trajectories[:,i] = prev

# visualize the results
figure(1)
hist(maxprev, 10)  
xlabel("Maximum prevalence")
ylabel("Frequency")

figure(2)
hist(durations,10) 
xlabel("Epidemic duration")
ylabel("Frequency")

figure(3)
for i in range(0,numruns):
#    scatter(range(timesteps), trajectories[:,i], s=5)
    plot(range(timesteps), trajectories[:,i])
xlabel("Time")
ylabel("Prevalence")

figure(4)
fill_between(range(timesteps), np.quantile(trajectories, .05, axis = 1), 
             np.quantile(trajectories, .95, axis = 1), facecolor='#7593C4')
fill_between(range(timesteps), np.quantile(trajectories, .25, axis = 1), 
             np.quantile(trajectories, .75, axis = 1), facecolor='#B8CDEF')
plot(range(timesteps), np.quantile(trajectories, .5, axis = 1), 'k',linewidth = 4)
xlabel("Time")
ylabel("Prevalence")


'''
Sample parameter space and explore behavior
'''

# Set up simulation parameters
nsamples = 100      # Number of LHS samples to draw
reruns = 2          # How many simulations to run on each parameter sample (to capture stochastic variation)
timesteps = 150     # Time steps to run each simulation

# Containers to hold the outputs
maxprev = np.zeros([1,nsamples*reruns])
durations = np.zeros([1,nsamples*reruns])
trajectories = np.zeros([timesteps,nsamples*reruns])

# Set up parameter array
params = npm.repmat(lhs(4, samples = nsamples),reruns,1) # Each row is a new parameter set
                                    # Note that lhs defaults to a range of (0,1) for each parameter

# Run the simulations!
for i in range(0,nsamples*reruns):
    p_i = params[i,0] # infection probability per contact
    p_r = params[i,1] # recovery probability
    p_0 = params[i,2] # probability of breaking tie if infected neighbor
    p_1 = params[i,3] # probability of regrowing tie if non-infected
    initialize()
    for j in range(1,timesteps):
        update()
    #maxprev.append(max(prev))
    #durations.append(sum(i > 0 for i in prev))
    maxprev[0,i] = max(prev)
    durations[0,i] = sum(i > 0 for i in prev)
    trajectories[:,i] = prev

maxprev = maxprev[0]
durations = durations[0]

# Visualize the results - UQ

# histograms
figure(1)
hist(maxprev, 10)  
xlabel("Maximum prevalence")
ylabel("Frequency")

figure(2)
hist(durations,10) 
xlabel("Epidemic duration")
ylabel("Frequency")

# scatterplot
figure(3)
scatter(durations, maxprev)
xlabel("Epidemic duration")
ylabel("Maximum prevalence")

# density plot
figure(4)
sns.kdeplot(durations, maxprev, shade=True, cut=5)
xlabel("Epidemic duration")
ylabel("Maximum prevalence")

# plot the model dynamics - all trajectories
figure(5)
for i in range(0,nsamples):
#    scatter(range(timesteps), trajectories[:,i], s=5)
    plot(range(timesteps), trajectories[:,i])
xlabel("Time")
ylabel("Prevalence")

# plot the model dynamics - quantiles
figure(6)
fill_between(range(timesteps), np.quantile(trajectories, .05, axis = 1), 
             np.quantile(trajectories, .95, axis = 1), facecolor='#7593C4')
fill_between(range(timesteps), np.quantile(trajectories, .25, axis = 1), 
             np.quantile(trajectories, .75, axis = 1), facecolor='#B8CDEF')
plot(range(timesteps), np.quantile(trajectories, .5, axis = 1), 'k',linewidth = 4)
xlabel("Time")
ylabel("Prevalence")



'''
Extra stuff
'''

#def observe():
#    global g, prev
#    cla()
#    nx.draw(g, cmap = cm.Wistia, vmin = 0, vmax = 1,
#            node_color = [g.node[i]['state'] for i in g.nodes],
#            pos = g.pos)
#
#pycxsimulator.GUI().start(func=[initialize, observe, update])
#
#
#epicurve = scatter(range(len(prev)), prev)
#xlabel("Time")
#ylabel("Prevalence")
#show(epicurve)

